/*
  k2917.c
  VME leaf driver for K2917 & K3922 
*/

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#ifdef MODULE
#include <linux/module.h>
#endif

#include <linux/types.h>
#include <asm/io.h>
#include <asm/system.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <asm/irq.h>
#include <asm/page.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <linux/sched.h>

#include <linux/signal.h>
#include <linux/ioport.h>
#include <linux/malloc.h>

#include "vmehb.h"

#include "k2917.h"

#define NTRY 10
#define NO_SIGNAL (!(current->signal.sig[0] & ~current->blocked.sig[0]))


struct k2917_struct {
  int use;             
  int major;
  int vmeminor;
  unsigned long base;
  unsigned int vme_size;
  volatile int int_counter;
  int crate;
};

static struct k2917_struct k2917_table = {
  0,
  K_MAJOR,
  0,
  K2917ADR,
  K2917SIZE,
  0,
  0
};


static struct wait_queue *k2917_wait_q;

struct k2917_list_action_struct {
  int use;
  int list_size;
  int *alist;
  int size;
  int *buf;
};

#define MAXLISTNUM 10
static struct k2917_list_action_struct k2917_la_table[MAXLISTNUM];

static int k2917_make_list_table( unsigned long );
static int k2917_delete_list_table( unsigned long );
static int k2917_list_action1( unsigned long );
static int k2917_list_action2( unsigned long );
static int k2917_list_action3( unsigned long );


static inline void _kin_write( int to, unsigned short data )
{
#ifdef MMAP
  unsigned short *ptr;
  ptr=(unsigned short *)(k2917_table.base+to); *ptr=data;
#else  
  vmehb_dev_write(k2917_table.vmeminor,k2917_table.base+to,(char *)&data,2);
#endif
}

static inline unsigned short _kin_read( int from )
{
#ifdef MMAP
  unsigned short *ptr;
  ptr=(unsigned short *)(k2917_table.base+from); 
  return *ptr;
#else
  unsigned short data;
  vmehb_dev_read(k2917_table.vmeminor,k2917_table.base+from,(char *)&data,2);
  return data;
#endif
}



static int k2917_int( int level, int resp )
{
#ifdef DEBUGINT
  unsigned short a=0;
#endif

  if(level==INTR_LEVEL && resp==INTR_VECTOR 
#ifdef DEBUGINT
     && ((a=_kin_read(CSR))&0x0200) 
#endif
     ){
    k2917_table.int_counter=1;
  }
  else{
#ifdef DEBUGINT
    printk("k2917: Error : Level %d : Resp %d : CSR %4hx\n",
	   level,resp,a);
#endif
    return 0;
  }
#ifdef DEBUGINT
  printk("k2917: Intr : L %d : R %d : CSR %4hx ICR %4hx IVR %4x\n",
	 level,resp,a,_kin_read(LAMICR),_kin_read(LAMIVR));
#endif
  wake_up_interruptible(&k2917_wait_q);
  return 1;
}

static int k2917_open( struct inode *inode, struct file *file )
{
  if(k2917_table.use==1){
    printk("k2917: Error : device busy\n");
    return -EBUSY;
  }
  _kin_write(CSR,0x1000); _kin_write(CSER,0); _kin_write(DOCR,0x8010);
  _kin_write(SCCR,0);     _kin_write(MTCR,0);

  _kin_write(LAMICR,0);   _kin_write(DONEICR,0);
  _kin_write(EMPTYICR,0); _kin_write(ABTICR,0);
  _kin_write(LAMIVR,0x0F);   _kin_write(DONEIVR,0x0F);
  _kin_write(EMPTYIVR,0x0F); _kin_write(ABTIVR,0x0F);

  _kin_write(AMR,0x79); _kin_write(CMAR,0x1FFF); _kin_write(CMR,0x80);

  request_vme_irq(INTR_LEVEL,INTR_VECTOR,k2917_int,"k2917");

  k2917_table.int_counter=0; k2917_table.use=1;

  printk("k2917: opened by \"%s\" (pid %i)\n",current->comm,current->pid);

#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
  return 0;
}

static int k2917_release( struct inode *inode, struct file *file )
{
  int i;
#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif
  _kin_write(LAMICR,0);   _kin_write(DONEICR,0);
  _kin_write(EMPTYICR,0); _kin_write(ABTICR,0);
  _kin_write(LAMIVR,0x0F);   _kin_write(DONEIVR,0x0F);
  _kin_write(EMPTYIVR,0x0F); _kin_write(ABTIVR,0x0F);

  free_vme_irq(INTR_LEVEL,INTR_VECTOR,k2917_int);

  /* free action list area */
  for(i=0;i<MAXLISTNUM;i++){
    if(k2917_la_table[i].use)
      k2917_delete_list_table((unsigned long)i);
  }

  k2917_table.int_counter=0; k2917_table.use=0;


  printk("k2917: closed by \"%s\" (pid %i)\n",current->comm,current->pid);

  return 0;
}

static int _kin_sst_ptro( unsigned short  naf, int ns, unsigned short *low,
			  unsigned short *high, int *q, int *x )
{
  volatile unsigned short status;
  int i;

#ifdef DEBUG
  printk("k2917:[PTRO] N=%2d A=%2d F=%2d NS=%d L:%4hx H:%4x\n",
	 NAFGetN(naf),NAFGetA(naf),NAFGetF(naf),ns,*low,*high);
#endif

  /* set CMAR to beginning */
  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(k2917_table.crate<<8)|((ns&0x1)<<1)|0x1);
  _kin_write(CMR,naf); /* _kin_write(CMR,0x80); */
  /* set CMAR to beginning and GO  */
  _kin_write(CMAR,0x1FFD);  _kin_write(CSR,0x01);

  (*q)=(*x)=0;
  do{
    i=-1;
    do{
      i++; status=_kin_read(CSR);
#ifdef DEBUG
      printk("k2917:[PTRO1]: %d %4hx\n",i,status);
#endif
    } while((status&0x8180)==0 && i<NTRY);
    if(i>=NTRY) return -ETIMEDOUT;
    if(status&0x0100){
      *low=_kin_read(DLR);
      if(ns==0) *high=_kin_read(DHR);
      status=_kin_read(CSR);
    }
#ifdef DEBUG
    printk("k2917:[PTRO2]: %4hx\n",status);
#endif
  }while((status&0x8080)==0);

  *q=(status&0x2)?0:1; *x=(status&0x4)?0:1;

#ifdef DEBUG
    printk("k2917:[PTRO3]: %4hx\n",status);
#endif
  if(status&0x2000) return -ETIMEDOUT;
  return 0;
}

static int _kin_sst_ptwo( unsigned short naf, int ns, unsigned short *low,
			  unsigned short *high, int *q, int *x )
{
  volatile unsigned short status;
  int i;
#ifdef DEBUG
  printk("k2917:[PTWO] N=%2d A=%2d F=%2d NS=%d L:%4hx H:%4x\n",
	 NAFGetN(naf),NAFGetA(naf),NAFGetF(naf),ns,*low,*high);
#endif
  /* set CMAR to beginning */
  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(k2917_table.crate<<8)|((ns&0x1)<<1)|0x1);
  _kin_write(CMR,naf);  /* _kin_write(CMR,0x80); */
  /* set CMAR to beginning and GO  */
  _kin_write(CMAR,0x1FFD);  _kin_write(CSR,0x21);

  do{
    i=-1;
    do{
      i++; status=_kin_read(CSR);
#ifdef DEBUG
      printk("k2917:[PTWO1]: %d %4hx\n",i,status);
#endif
    }while((status&0x8180)==0 && i<NTRY);
    if(i>=NTRY) return -ETIMEDOUT;
    if(status&0x0100){
      _kin_write(DLR,*low);
      if(ns==0) _kin_write(DHR,*high);
      status=_kin_read(CSR);
    }
#ifdef DEBUG
    printk("k2917:[PTWO2]: %4hx\n",status);
#endif
  }while((status&0x8080)==0);

  *q=(status&0x2)?0:1; *x=(status&0x4)?0:1;

#ifdef DEBUG
    printk("k2917:[PTWO3]: %4hx\n",status);
#endif
  if(status&0x2000) return -ETIMEDOUT;
  return 0;
}

static int _kin_sst_ptco( unsigned short naf, int *q, int *x )
{
  volatile unsigned short status;
  int i;

#ifdef DEBUG
  printk("k2917:[PTCO] N=%2d A=%2d F=%2d\n",
	 NAFGetN(naf),NAFGetA(naf),NAFGetF(naf));
#endif

  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(k2917_table.crate<<8)|0x1);
  _kin_write(CMR,naf);
  /*  _kin_write(CMR,0x80); */
  /* set CMAR to beginning and GO */ 
  _kin_write(CMAR,0x1FFD);  _kin_write(CSR,0x01);

  i=-1;
  do{
    i++; status=_kin_read(CSR);
#ifdef DEBUG
    printk("k2917:[PTCO1]: %d %4hx\n",i,status);
#endif
  }while((status&0x8180)==0 && i<NTRY);

  if(i>=NTRY) return -ETIMEDOUT;

  *q=(status&0x2)?0:1; *x=(status&0x4)?0:1;

#ifdef DEBUG
  printk("k2917:[PTCO2]: %d %4hx\n",i,status);
#endif
  if(status&0x2000) return -ETIMEDOUT;
  return 0;
}

static int k2917_ioctl( struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg )
{
  unsigned long timeout=HZ;
  int retval=1,size;
  unsigned short low,high,naf;
  struct k2917_single single;
  int q,x,st;
  unsigned flags;

#ifdef DEBUG
  printk("k2917: (ioctl) cmd=%8x arg=%8lx\n",cmd,arg);
  printk("TYPE=%x : NR=%d DIR=%x : SIZE=%d\n",
	 _IOC_TYPE(cmd),_IOC_NR(cmd),_IOC_DIR(cmd),_IOC_SIZE(cmd));
#endif
  /* command arg. etc check */
  size=_IOC_SIZE(cmd);
  if(!k2917_table.use)                return -EBADF;
  if(_IOC_TYPE(cmd)!=K2917_IOC_MAGIC) return -EINVAL;
  if(_IOC_NR(cmd)>K2917_IOC_MAXNR)    return -EINVAL;
  if(_IOC_DIR(cmd) & _IOC_READ)
    retval=access_ok(VERIFY_WRITE,(void *)arg,size);
  else if(_IOC_DIR(cmd) & _IOC_WRITE)
    retval=access_ok(VERIFY_READ,(void *)arg,size);
  if(!retval)                         return -EFAULT;

  retval=0;
  switch(cmd){
  case K2917_IOC_RESET:  /* Reset Controller */
    save_flags(flags); cli(); 
    _kin_write(CSR,0x1000); 
    _kin_write(CMAR,0x1FFF); _kin_write(CMR,0x80);
    restore_flags(flags);
    break;
  case K2917_IOC_EI:     /* Enable Interrupt */
    /* N=30 A=13 F=17 W24=mask  (Write LAM Mask) */
    low=(unsigned short)(arg&0xFFFF); 
    high=(unsigned short)((arg>>16)&0x00FF);
    naf=NAF(30,13,17);
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,0,&low,&high,&q,&x);
    restore_flags(flags);
#ifdef DEBUGINT
    printk("k2917: IOC_EI : Mask=%8x Status=%d\n",
	   (int)(((high<<16)&0xFF0000)|(low&0xFFFF)),st);
#endif
    if(st) return st;
    /* N=30 A=0 F=1 R16 (Read Status) */ 
    naf=NAF(30,0,1);
    save_flags(flags); cli();
    st=_kin_sst_ptro(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    /* N=30 A=0 F=17 W16 (Write Status) */ 
    low=(low|0x100)&K3922SR_WMASK;
    naf=NAF(30,0,17);
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    save_flags(flags); cli();
    _kin_write(LAMIVR,INTR_VECTOR); 
    _kin_write(LAMICR,0x18+INTR_LEVEL);
    restore_flags(flags);
    break;
  case K2917_IOC_DI:     /* Disable Interrupt */
    /* N=30 A=13 F=17 W24 (Write LAM Mask) */ 
    naf=NAF(30,13,17); low=high=0;
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,0,&low,&high,&q,&x);
    restore_flags(flags);
#ifdef DEBUGINT
    printk("k2917: IOC_DI : Mask=%8x Status=%d\n",
	   (int)(((high<<16)&0xFF0000)|(low&0xFFFF)),st);
#endif
    if(st) return st;
    /* N=30 A=0 F=1 R16 (Read Status) */
    naf=NAF(30,0,1);
    save_flags(flags); cli();
    st=_kin_sst_ptro(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    /* N=30 A=0 F=17 W16 (Write Status) */ 
    low=low&K3922SR_WMASK&(~0x100);
    naf=NAF(30,0,17);
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    _kin_write(LAMICR,0); _kin_write(LAMIVR,0x0F);
    break;
  case K2917_IOC_SETCR:  /* Set Crate Number */
    if(arg>=0 && arg<8)
      k2917_table.crate=(int)arg;
    else
      return -EINVAL;
    break;
  case K2917_IOC_WLAM:   /* Wait LAM Intterupt */
    save_flags(flags); cli();
    _kin_write(LAMICR,INTR_LEVEL+0x18);
    if(!k2917_table.int_counter){
      if(arg) timeout=arg*HZ;
#ifdef DEBUGINT
      printk("k2917: Timeout %ld Sleep On Jiffies=%lu\n",timeout,jiffies);
#endif
      interruptible_sleep_on_timeout(&k2917_wait_q,timeout);
    }
#ifdef DEBUGINT
    restore_flags(flags);
    printk("k2917: IntCounter=%d Signal=%lx BSignal=%lx Jiffies=%lu\n",
	   k2917_table.int_counter,current->signal.sig[0],
	   current->blocked.sig[0],jiffies);
    save_flags(flags); cli();
#endif
    /*    if(k2917_table.int_counter && NO_SIGNAL){ */
    if(k2917_table.int_counter){
#ifdef DEBUGINT
      printk("k2917: Interrupt :\n");
#endif
      k2917_table.int_counter=0; 
    }
    else{
#ifdef DEBUGINT
      printk("k2917: timeout\n");
#endif
      retval=-ETIMEDOUT;
    }
    restore_flags(flags);
    break;
  case K2917_IOC_CAMR16: /* CAMAC 16 Bits Read */
    if(__copy_from_user(&single,(void *)arg,size))
      return -EFAULT;
    /*    naf=single.naf&0xFFFF; */
    naf=single.naf;
    save_flags(flags); cli();
    st=_kin_sst_ptro(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    /*    single.data=low&0xFFFF;  */
    single.data=low; 
    single.q=q; single.x=x;
    if(__copy_to_user((void *)arg,&single,size))
      retval=-EFAULT;
    break;
  case K2917_IOC_CAMR24: /* CAMAC 24 Bits Read */
    if(__copy_from_user(&single,(void *)arg,size))
      return -EFAULT;
    /*    naf=single.naf&0xFFFF; */
    naf=single.naf;
    save_flags(flags); cli();
    st=_kin_sst_ptro(naf,0,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    /*    single.data=((high&0xFF)<<16)|(low&0xFFFF);  */
    single.data=(high<<16)|low; 
    single.q=q; single.x=x;
    if(__copy_to_user((void *)arg,&single,size))
      retval=-EFAULT;
    break;
  case K2917_IOC_CAMW16: /* CAMAC 16 Bits Write */
    if(__copy_from_user(&single,(void *)arg,size))
      return -EFAULT;
    /*    naf=single.naf&0xFFFF; */
    naf=single.naf;
    low=single.data;
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,1,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    single.q=q; single.x=x;
    if(__copy_to_user((void *)arg,&single,size))
      retval=-EFAULT;
    break;
  case K2917_IOC_CAMW24: /* CAMAC 24 Bits Write */
    if(__copy_from_user(&single,(void *)arg,size))
      return -EFAULT;
    /* naf=single.naf&0xFFFF; */
    naf=single.naf;
    low=single.data&0xFFFF; high=(single.data>>16)&0xFF;
    save_flags(flags); cli();
    st=_kin_sst_ptwo(naf,0,&low,&high,&q,&x);
    restore_flags(flags);
    if(st) return st;
    single.q=q; single.x=x;
    if(__copy_to_user((void *)arg,&single,size))
      retval=-EFAULT;
    break;
  case K2917_IOC_CAMC:   /* CAMAC Control */
    if(__copy_from_user(&single,(void *)arg,size))
      return -EFAULT;
    /* naf=single.naf&0xFFFF; */
    naf=single.naf;
    save_flags(flags); cli();
    st=_kin_sst_ptco(naf,&q,&x);
    restore_flags(flags);
    single.q=q; single.x=x;
    if(__copy_to_user((void *)arg,&single,size))
      retval=-EFAULT;
    break;
  case K2917_IOC_ALREG:
    retval=k2917_make_list_table(arg);
    break;
  case K2917_IOC_ALUNREG:
    retval=k2917_delete_list_table(arg);
    break;
  case K2917_IOC_EXEC_LA1:
    retval=k2917_list_action1(arg);
    break;
  case K2917_IOC_EXEC_LA2:
    retval=k2917_list_action2(arg);
    break;
  case K2917_IOC_EXEC_LA3:
    retval=k2917_list_action3(arg);
    break;
  default:
    retval=-EINVAL;
    break;
  }
  return retval;
}

static struct file_operations k2917_fops = {
  NULL,          /* lseek        */
  NULL,          /* read         */
  NULL,          /* write        */
  NULL,          /* readdir      */
  NULL,          /* pol          */
  k2917_ioctl,   /* ioctl        */
  NULL,          /* mmap         */
  k2917_open,    /* open         */
  NULL,          /* flush        */ 
  k2917_release, /* release      */  
  NULL,          /* fsync        */
  NULL,          /* fasync       */
  NULL,          /* media change */
  NULL,          /* revailodate  */
  NULL           /* lock         */
};

/*
static int k2917_defint( int level, int resp )
{
  printk("k2917: dummy handler L:%d V:%d\n",level,resp);
  return 0;
}
*/


#ifdef MODULE
#ifdef MMAP
static unsigned int mmap_base,mmap_offset,mmap_length;
#endif

int init_module( void )
{
#ifdef MMAP
  unsigned long base;
#endif
  /*
  request_vme_irq(INTR_LEVEL,0x0F,k2917_defint,"k2917_def");
  */
  if(register_chrdev(k2917_table.major,"k2917",&k2917_fops)){
    printk("k2917: Error : cannot register to major device %d\n",
	   k2917_table.major);
    goto error1;
  }

#ifdef MMAP
  mmap_offset=K2917ADR;
  mmap_length=K2917SIZE+mmap_offset%PAGE_SIZE;
  mmap_base=mmap_offset/PAGE_SIZE*PAGE_SIZE;
  printk("--------------------------------------------------------\n");
  printk("k2917: MMAP version Ofs: %8x Siz: %8x Base: %8x\n",
	 mmap_offset,mmap_length,mmap_base);
  if((base=vmehb_dev_mmap(k2917_table.vmeminor,mmap_base,mmap_length))==0){
    printk("k2917: Error : cannot mmap device %x-%x\n",
	   mmap_base,mmap_base+mmap_length-1);
    goto error2;
  }
  k2917_table.base=base+(mmap_offset%PAGE_SIZE);
  printk("k2917: MMAP version VMEaddr.:%x Base:%lx\n",
	 K2917ADR,k2917_table.base);
#else
  printk("--------------------------------------------------------\n");
  printk("k2917: RWA version VMEaddr.:%lx\n",k2917_table.base);
#endif

  _kin_write(CSR,0x1000); _kin_write(CSER,0); _kin_write(DOCR,0x8010);
  _kin_write(SCCR,0);     _kin_write(MTCR,0);

  _kin_write(LAMICR,0);   _kin_write(DONEICR,0);
  _kin_write(EMPTYICR,0); _kin_write(ABTICR,0);
  _kin_write(LAMIVR,0x0F);   _kin_write(DONEIVR,0x0F);
  _kin_write(EMPTYIVR,0x0F); _kin_write(ABTIVR,0x0F);

  _kin_write(AMR,0x79); _kin_write(CMAR,0x1FFF); _kin_write(CMR,0x80);
  return 0;

#ifdef MMAP
 error2:      
  unregister_chrdev(k2917_table.major,"k2917");
#endif
 error1:
  return -ENODEV;
}

void cleanup_module( void )
{
  /*
  free_vme_irq(INTR_LEVEL,0x0F,k2917_defint);
  */
#ifdef MMAP
  vmehb_dev_unmap(k2917_table.vmeminor,mmap_base,mmap_length);
#endif
  unregister_chrdev(k2917_table.major,"k2917");
  printk("k2917: removed\n");
}

#endif


static int k2917_make_list_table( unsigned long arg )
     /* arg -- pointer to struct k2917_al_struct (user space) */
{
  int retval,num,size,bufsize;
  int *list,*buf;
  struct k2917_al_struct al;

  /* check user space */
  retval=__copy_from_user(&al,(void *)arg,sizeof(struct k2917_al_struct));
  if(retval) return -EFAULT;
  size=al.size; bufsize=al.buffer_size;
  retval=access_ok(VERIFY_READ,al.action_list,sizeof(int)*size);
  if(!retval) return -EFAULT;

  if((num=al.al_num)<0 || num>=MAXLISTNUM)
    return -EINVAL;

  if(k2917_la_table[num].use){
#ifdef DEBUGLIST
    printk("k2917: Table %d is already in use\n",num);
#endif
    return -EINVAL;
  }
  /* memory allocation */
  if((list=kmalloc(sizeof(int)*size,GFP_KERNEL))==0){
#ifdef DEBUGLIST
    printk("k2917: kmalloc(1) fail\n");
#endif
    return -EFAULT;
  }
  if((buf=kmalloc(sizeof(int)*bufsize,GFP_KERNEL))==0){
#ifdef DEBUGLIST
    printk("k2917: kmalloc(2) fail\n");
#endif
    kfree(list);
    return -EFAULT;
  }
  retval=__copy_from_user(list,al.action_list,sizeof(int)*size);
  if(retval){
#ifdef DEBUGLIST
    printk("k2917: Action list copy fails\n");
#endif
    kfree(list); kfree(buf);
    return -EFAULT;
  }
  k2917_la_table[num].use=1;
  k2917_la_table[num].list_size=size;
  k2917_la_table[num].alist=list;
  k2917_la_table[num].size=bufsize;
  k2917_la_table[num].buf=buf;
#ifdef DEBUGLIST
  printk("k2917: IOC_ALREG : Num=%d LS=%d BS~%d\n",
	 num,k2917_la_table[num].list_size,k2917_la_table[num].size);
  printk("k2917: === list [%8lx] : buffer [%8lx] ===\n",
	 (unsigned long)k2917_la_table[num].alist,
	 (unsigned long)k2917_la_table[num].buf);
#endif
  return 0;
}

static int k2917_delete_list_table( unsigned long arg )
     /* arg -- table number */
{
  int num=(int)arg;
#ifdef DEBUGLIST
  printk("k2917: (delete_table) Num=%d\n",num);
#endif
  if(num<0 || num>=MAXLISTNUM)
    return -EINVAL;
  if(!k2917_la_table[num].use){
#ifdef DEBUGLIST
    printk("k2917: (delete_table) Table %d is not in use\n",num);
#endif
    return -EINVAL;
  }

  if(k2917_la_table[num].alist)
    kfree(k2917_la_table[num].alist);
  if(k2917_la_table[num].buf)
    kfree(k2917_la_table[num].buf);
  k2917_la_table[num].use=0;
  k2917_la_table[num].list_size=0;
  k2917_la_table[num].alist=0;
  k2917_la_table[num].size=0;
  k2917_la_table[num].buf=0;

  return 0;
}

static int k2917_list_action1( unsigned long arg )
     /* arg -- pointer to struct k2917_ad_struct (user space) */
{
  struct k2917_ad_struct ad;
  int retval,num,i,j;
  int umax,nlist,wsize;
  int *ubuffer,*kbuffer,*list;
  int n,a,q,x; 
  unsigned short naf,st,low;
  unsigned flags;

  /* check user space */
  retval=__copy_from_user(&ad,(void *)arg,sizeof(struct k2917_ad_struct));
  if(retval) return -EFAULT;

  umax=ad.maxsize; ubuffer=ad.buffer; num=ad.al_num;
#ifdef DEBUGLIST
  printk("k2917: (action1) : Num=%d MaxBufSiz=%d Buf [%8lx]\n",
	 num,umax,(unsigned long)ubuffer);
#endif
  retval=access_ok(VERIFY_WRITE,ubuffer,sizeof(int)*umax);
  if(!retval) return -EFAULT;
  if(num<0 || num>=MAXLISTNUM){
#ifdef DEBUGLIST
    printk("k2917: (action1) : Invalid List Num %d\n",num);
#endif
    return -EINVAL;
  }
  kbuffer=k2917_la_table[num].buf;
  list=k2917_la_table[num].alist;
  nlist=k2917_la_table[num].list_size;

  if(umax>k2917_la_table[num].size)
    umax=k2917_la_table[num].size;

  /* Action starts here */
  save_flags(flags); cli();
  for(i=0,wsize=0; i<nlist&&wsize<umax; i++){
    naf=list[i]&NAFMASK; 
    n=NAFGetN(list[i]); a=NAFGetA(list[i]);
    _kin_write(CMAR,0x1FFD);
    _kin_write(CMR,(k2917_table.crate<<8)|0x3);
    _kin_write(CMR,naf);
    /*    _kin_write(CMR,0x80); */
    _kin_write(CMAR,0x1FFD); _kin_write(CSR,0x01);
    j=-1;
    do{
      st=_kin_read(CSR); j++;
    } while((st&0x8180)==0 && j<NTRY);
    if(st&0x0100){
      low=_kin_read(DLR);
      st=_kin_read(CSR);
      q=(st&0x2)?0:1; x=(st%0x4)?0:1;
      kbuffer[wsize++]=LDencode(q,x,n,a,low);
    }
  }
  restore_flags(flags);
  /* Action ends */
  ad.size=wsize;
#ifdef DEBUGLIST
  printk("k2917: (action1) Wsize=%d\n",ad.size);
#endif
  if(__copy_to_user(ubuffer,kbuffer,sizeof(int)*wsize))
    return -EFAULT;
  if(__copy_to_user((void *)arg,&ad,sizeof(struct k2917_ad_struct)))
    return -EFAULT;

  return 0;
}

static int k2917_list_action2( unsigned long arg )
     /* arg -- pointer to struct k2917_ad_struct (user space) */
{
  struct k2917_ad_struct ad;
  int retval,num,i,j;
  int umax,nlist,wsize;
  int *ubuffer,*kbuffer,*list;
  int n,a,q,x,qs; 
  unsigned short naf,st,low;
  unsigned flags;

  /* check user space */
  retval=__copy_from_user(&ad,(void *)arg,sizeof(struct k2917_ad_struct));
  if(retval) return -EFAULT;
  umax=ad.maxsize; ubuffer=ad.buffer; num=ad.al_num;
#ifdef DEBUGLIST
  printk("k2917: (action2) : Num=%d MaxBufSiz=%d Buf [%8lx]\n",
	 num,umax,(unsigned long)ubuffer);
#endif
  retval=access_ok(VERIFY_WRITE,ubuffer,sizeof(int)*umax);
  if(!retval) return -EFAULT;
  if(num<0 || num>=MAXLISTNUM){
#ifdef DEBUGLIST
    printk("k2917: (action2) : Invalid List Num %d\n",num);
#endif
    return -EINVAL;
  }
  kbuffer=k2917_la_table[num].buf;
  list=k2917_la_table[num].alist;
  nlist=k2917_la_table[num].list_size;

  if(umax>k2917_la_table[num].size)
    umax=k2917_la_table[num].size;

  /* Action starts here */
  save_flags(flags); cli();
  for(i=0,wsize=0; i<nlist&&wsize<umax; i++){
    naf=list[i]&NAFMASK; 
    n=NAFGetN(list[i]); a=NAFGetA(list[i]); qs=list[i]&QS_READ;
    if(qs){
      _kin_write(CMAR,0x1FFB);
      _kin_write(CMR,(k2917_table.crate<<8)|0x23); 
      _kin_write(CMR,0x07FF);
      /*      _kin_write(CMR,0x80); */
      _kin_write(CMAR,0x1FFB); _kin_write(CSR,0x01);
      do{
	j=-1;
	do{
	  st=_kin_read(CSR); j++;
	} while((st&0x8180)==0 && j<NTRY);
	if(j>=NTRY) st|=0x2000;
	if(st&0x0100){
	  low=_kin_read(DLR);
	  st=_kin_read(CSR);
	  q=(st&0x2)?0:1;
	  if(q){
	    x=(st&0x4)?0:1;
	    kbuffer[wsize++]=LDencode(q,x,n,a,low);
	  }
	}
      } while(st&0xE080 && wsize<umax); 
    }
    else{
      _kin_write(CMAR,0x1FFD);
      _kin_write(CMR,(k2917_table.crate<<8)|0x3);
      _kin_write(CMR,naf);
      /*      _kin_write(CMR,0x80); */
      _kin_write(CMAR,0x1FFD); _kin_write(CSR,0x01);
      j=-1;
      do{
	st=_kin_read(CSR); j++;
      } while((st&0x8180)==0 && j<NTRY);
      if(st&0x0100){
	low=_kin_read(DLR);
	st=_kin_read(CSR);
	q=(st&0x2)?0:1;
	if(q){
	  x=(st%0x4)?0:1;
	  kbuffer[wsize++]=LDencode(q,x,n,a,low);
	}
      }
    }
  }
  restore_flags(flags);
  /* Action ends */
  ad.size=wsize;
#ifdef DEBUGLIST
  printk("k2917: (action2) Wsize=%d\n",ad.size);
#endif
  if(__copy_to_user(ubuffer,kbuffer,sizeof(int)*wsize))
    return -EFAULT;
  if(__copy_to_user((void *)arg,&ad,sizeof(struct k2917_ad_struct)))
    return -EFAULT;

  return 0;
}

static int k2917_list_action3( unsigned long arg )
     /* arg -- pointer to struct k2917_ad_struct (user space) */
{
  struct k2917_ad_struct ad;
  int retval,num,i,j;
  int umax,nlist,wsize;
  int *ubuffer,*kbuffer,*list;
  int n,a,q,x,qs,as,am,f,cset; 
  unsigned short naf,st,low;
  unsigned flags;

  /* check user space */
  retval=__copy_from_user(&ad,(void *)arg,sizeof(struct k2917_ad_struct));
  if(retval) return -EFAULT;
  umax=ad.maxsize; ubuffer=ad.buffer; num=ad.al_num;
#ifdef DEBUGLIST
  printk("k2917: (action3) : Num=%d MaxBufSiz=%d Buf [%8lx]\n",
	 num,umax,(unsigned long)ubuffer);
#endif
  retval=access_ok(VERIFY_WRITE,ubuffer,sizeof(int)*umax);
  if(!retval) return -EFAULT;
  if(num<0 || num>=MAXLISTNUM){
#ifdef DEBUGLIST
    printk("k2917: (action3) : Invalid List Num %d\n",num);
#endif
    return -EINVAL;
  }
  kbuffer=k2917_la_table[num].buf;
  list=k2917_la_table[num].alist;
  nlist=k2917_la_table[num].list_size;

  if(umax>k2917_la_table[num].size)
    umax=k2917_la_table[num].size;

  /* Action starts here */
  save_flags(flags); cli();
  for(i=0,wsize=0; i<nlist&&wsize<umax; i++){
    n=NAFGetN(list[i]); a=NAFGetA(list[i]);
    qs=list[i]&QS_READ; as=ASCHK(list[i]);
    naf=list[i]&NAFMASK;
    if(as){
      q=1; f=NAFGetF(list[i]); am=ASGetAM(list[i]); cset=0;
      while(q==1 && a<am && wsize<umax){
	naf=NAF(n,a,f);
	if(!cset){
	  _kin_write(CMAR,0x1FFD);
	  _kin_write(CMR,(k2917_table.crate<<8)|0x3);
	  cset=1;
	}
	else{
	  _kin_write(CMAR,0x1FFE);
	}
	_kin_write(CMR,naf);
	/* _kin_write(CMR,0x80); */
	_kin_write(CMAR,0x1FFD); _kin_write(CSR,0x01);
	j=-1;
	do{
	  st=_kin_read(CSR); j++;
	} while((st&0x8180)==0 && j<NTRY);
	if(st&0x0100){
	  low=_kin_read(DLR);
	  q=(st&0x2)?0:1;
	  if(q){
	    x=(st&0x4)?0:1;
	    kbuffer[wsize++]=LDencode(q,x,n,a,low);
	  }
	  else{
	    break;
	  }
	}
	else{
	  q=0;
	}
	a++;
      }
    }
    else if(qs){
      _kin_write(CMAR,0x1FFB);
      _kin_write(CMR,(k2917_table.crate<<8)|0x23); 
      _kin_write(CMR,0x07FF);
      /*      _kin_write(CMR,0x80); */
      _kin_write(CMAR,0x1FFB); _kin_write(CSR,0x01);
      do{
        j=-1;
        do{
          st=_kin_read(CSR); j++;
        } while((st&0x8180)==0 && j<NTRY);
        if(j>=NTRY) st|=0x2000;
        if(st&0x0100){
          low=_kin_read(DLR);
          st=_kin_read(CSR);
          q=(st&0x2)?0:1;
          if(q){
            x=(st&0x4)?0:1;
	    kbuffer[wsize++]=LDencode(q,x,n,a,low);
	  }
	}
      } while(st&0xE080 && wsize<umax); 
    }
    else{
      _kin_write(CMAR,0x1FFD);
      _kin_write(CMR,(k2917_table.crate<<8)|0x3);
      _kin_write(CMR,naf);
      /*      _kin_write(CMR,0x80); */
      _kin_write(CMAR,0x1FFD); _kin_write(CSR,0x01);
      j=-1;
      do{
        st=_kin_read(CSR); j++;
      } while((st&0x8180)==0 && j<NTRY);
      if(st&0x0100){
        low=_kin_read(DLR);
        st=_kin_read(CSR);
        q=(st&0x2)?0:1;
        if(q){
          x=(st%0x4)?0:1;
          kbuffer[wsize++]=LDencode(q,x,n,a,low);
	}
      }
    }
   }
  restore_flags(flags);
  /* Action ends */
  ad.size=wsize;
#ifdef DEBUGLIST
  printk("k2917: (action3) Wsize=%d\n",ad.size);
#endif
  if(__copy_to_user(ubuffer,kbuffer,sizeof(int)*wsize))
    return -EFAULT;
  if(__copy_to_user((void *)arg,&ad,sizeof(struct k2917_ad_struct)))
    return -EFAULT;

  return 0;
}
